{*******************************************************}
{                                                       }
{       RichView                                        }
{       Demo: Dual unicode editor with synchronized     }
{       scrolling and highlighting.                     }
{                                                       }
{       Copyright (c) Sergey Tkachenko                  }
{       svt@trichview.com                               }
{       http://www.trichview.com                        }
{                                                       }
{*******************************************************}

{ Limitation: this demo does not work properly with tables.
  Compile it with RVDONOTUSETABLES compiler define (see RV_Defs.inc).
  Note 1: when running in Delphi, an exception is possible in RVStyle1DrawParaBack.
    This exception is handled in the component and does not dangerous.
  Note 2: possible optimization: in RichViewEditCaretMove, redraw an unfocused
    editor only if its HltItemNo is changed.
}

unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, RVStyle, RVScroll, RichView, RVEdit, CRVData, StdCtrls;

type
  TForm1 = class(TForm)
    RVStyle1: TRVStyle;
    Panel1: TPanel;
    Panel2: TPanel;
    RichViewEdit1: TRichViewEdit;
    RichViewEdit2: TRichViewEdit;
    Splitter1: TSplitter;
    Button1: TButton;
    Button2: TButton;
    Label1: TLabel;
    CheckBox1: TCheckBox;
    procedure RichViewEdit1VScrolled(Sender: TObject);
    procedure RichViewEdit2VScrolled(Sender: TObject);
    procedure RVStyle1DrawParaBack(Sender: TRVStyle; Canvas: TCanvas;
      ParaNo: Integer; ARect: TRect; var DoDefault: Boolean);
    procedure RichViewEditCaretMove(Sender: TObject);
    procedure RichViewEditEnter(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure CheckBox1Click(Sender: TObject);
  private
    { Private declarations }
    NoUpdate: Boolean;
    FocusedEditor: TObject;
    HltItemNo1, HltItemNo2: Integer;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

{ Loading default files }
procedure TForm1.FormCreate(Sender: TObject);
begin
  RichViewEdit1.LoadRVF(ExtractFilePath(Application.ExeName)+'en.rvf');
  RichViewEdit2.LoadRVF(ExtractFilePath(Application.ExeName)+'ru.rvf');
  RichViewEdit1.Format;
  RichViewEdit2.Format;
end;

{============================ Scroll synchronization ==========================}

{ Returns the index of the first item of the same paragraph as ItemNo }
function GetFirstParagraphItemNo(rve: TCustomRichViewEdit; ItemNo: Integer): Integer;
begin
  while not rve.IsFromNewLine(ItemNo) do
    dec(ItemNo);
  Result := ItemNo;
end;

{ Returns the index of the last item of the same paragraph as ItemNo }
function GetLastParagraphItemNo(rve: TCustomRichViewEdit; ItemNo: Integer): Integer;
begin
  while (ItemNo+1<rve.ItemCount-1) and not rve.IsFromNewLine(ItemNo+1) do
    inc(ItemNo);
  Result := ItemNo;
end;

{ Returns the index of paragraph (more exactly - of paragraph section }
function GetParagraphNo(rve: TCustomRichViewEdit; ItemNo: Integer): Integer;
var i: Integer;
begin
  Result := 0;
  for i := 0 to ItemNo do
    if rve.IsFromNewLine(i) then
      inc(Result);
end;

{ Returns the index of the first item of the specified paragraph }  
function GetFirstItemNoInParagraph(rve: TCustomRichViewEdit; ParagraphNo: Integer): Integer;
var i: Integer;
begin
  Result := -1;
  for i := 0 to rve.ItemCount-1 do
    if rve.IsFromNewLine(i) then begin
      dec(ParagraphNo);
      if ParagraphNo=0 then begin
        Result := i;
        exit;
      end;
    end;
end;

{ Scrolls to the paragraph with index ParagraphNo, shifting the result by Delta }
procedure ScrollToParagraphNo(rve: TCustomRichViewEdit; ParagraphNo, Delta: Integer);
var ItemNo, x,y: Integer;
begin
  ItemNo := GetFirstItemNoInParagraph(rve, ParagraphNo);
  if ItemNo<0 then
    exit;
  rve.GetItemCoords(ItemNo, x, y);
  rve.ScrollTo(y+Delta);
end;

{ Scrolls rve2 to the same position as rve1 }
procedure SyncPosition(rve1, rve2: TCustomRichViewEdit);
var ItemNo, ParagraphNo, x, y: Integer;
begin
  ItemNo := rve1.FirstItemVisible;
  if ItemNo<0 then
    exit;
  ItemNo := GetFirstParagraphItemNo(rve1, ItemNo);
  rve1.GetItemClientCoords(ItemNo, x, y);
  ParagraphNo := GetParagraphNo(rve1, ItemNo);
  ScrollToParagraphNo(rve2, ParagraphNo, -y);
end;

{ Sync 2 by 1 }
procedure TForm1.RichViewEdit1VScrolled(Sender: TObject);
begin
  if NoUpdate then
    exit;
  NoUpdate := True;
  SyncPosition(RichViewEdit1, RichViewEdit2);
  NoUpdate := False;
end;

{ Sync 1 by 2 }
procedure TForm1.RichViewEdit2VScrolled(Sender: TObject);
begin
  if NoUpdate then
    exit;
  NoUpdate := True;
  SyncPosition(RichViewEdit2, RichViewEdit1);
  NoUpdate := False;
end;

{========================== Paragraph Highlight ===============================}

{ Drawing paragraph background for both editors
  For focused editor, highlight the current paragraph. For another editor,
  highlights the paragraph containing item HltItemNo1/HltItemNo2. }
procedure TForm1.RVStyle1DrawParaBack(Sender: TRVStyle; Canvas: TCanvas;
  ParaNo: Integer; ARect: TRect; var DoDefault: Boolean);
var i, j: Integer;
    MakeRed: Boolean;
begin
  if not CheckBox1.Checked then
    exit;
  i := Sender.ItemNo;
  if Sender.RVData=RichViewEdit1.RVData then
    if FocusedEditor = RichViewEdit1 then
      j := RichViewEdit1.CurItemNo
    else
      j := HltItemNo1
  else
    if FocusedEditor = RichViewEdit2 then
      j := RichViewEdit2.CurItemNo
    else
      j := HltItemNo2;
  repeat
    MakeRed := i=j;
    if MakeRed then
      break;
    dec(i);
  until TCustomRVData(Sender.RVData).IsFromNewLine(i+1);
  if MakeRed then begin
    Canvas.Brush.Style := bsSolid;
    Canvas.Brush.Color := $CCCCFF;
    Canvas.FillRect(ARect);
  end;
end;

{ Updating HltItemNo for unfocused editor (rve2) when moving caret in
  the focused editor (rv1) }
procedure DoOnCaretMove(rve1, rve2: TCustomRichViewEdit;
  var HltItemNo: Integer);
var ItemNo, ParagraphNo: Integer;
begin
  ItemNo := rve1.CurItemNo;
  if ItemNo<0 then begin
    HltItemNo := -1;
    exit;
  end;
  ParagraphNo := GetParagraphNo(rve1, ItemNo);
  HltItemNo := GetFirstItemNoInParagraph(rve2, ParagraphNo);
  if HltItemNo>=0 then
    HltItemNo := GetLastParagraphItemNo(rve2, HltItemNo);
end;

{ When caret is moved... }
procedure TForm1.RichViewEditCaretMove(Sender: TObject);
begin
  if not CheckBox1.Checked then
    exit;
  if FocusedEditor = RichViewEdit2 then
    DoOnCaretMove(RichViewEdit2, RichViewEdit1, HltItemNo1)
  else
    DoOnCaretMove(RichViewEdit1, RichViewEdit2, HltItemNo2);
  RichViewEdit1.Invalidate;
  RichViewEdit2.Invalidate
end;

{ Updating focused editor reference }
procedure TForm1.RichViewEditEnter(Sender: TObject);
begin
  FocusedEditor := Sender;
end;

{=========================== Word Selection ===================================}

{ Selecting previous word in the ItemNo-th item of rve. Offs - caret position.
  Moved - if the caret was already moved }
function SelectPrevWordInItem(rve: TCustomRichViewEdit; ItemNo, Offs: Integer;
  Moved: Boolean): Boolean;
var ws: WideString;
    WordEndOffs: Integer;
begin
  Result := False;
  if rve.GetItemStyle(ItemNo)<0 then
    exit;
  ws := rve.GetItemTextW(ItemNo);
  dec(Offs);
  // skipping delimiters
  while (Offs>0) and rve.RVData.IsDelimiterW(Word(ws[Offs])) do begin
    dec(Offs);
    Moved := True;
  end;
  if Offs=0 then
    exit;
  // if the caret was not moved, skipping the current word and delimiters before it    
  if not Moved then begin
    while (Offs>0) and not rve.RVData.IsDelimiterW(Word(ws[Offs])) do
      dec(Offs);
    if Offs=0 then
      exit;
    while (Offs>0) and rve.RVData.IsDelimiterW(Word(ws[Offs])) do
      dec(Offs);
    if Offs=0 then
      exit;
  end;
  // we are at the end of word, finding its beginning
  WordEndOffs := Offs;
  while (Offs>0) and not rve.RVData.IsDelimiterW(Word(ws[Offs])) do
    dec(Offs);
  rve.SetSelectionBounds(ItemNo, Offs+1, ItemNo, WordEndOffs+1);
  rve.Invalidate;
  Result := True;
end;

{ Selecting the previous word in rve }
function SelectPrevWord(rve: TCustomRichViewEdit): Boolean;
var ItemNo, Offs: Integer;
    Moved: Boolean;
begin
  Result := False;
  ItemNo := rve.CurItemNo;
  Offs   := rve.OffsetInCurItem;
  Moved := False;
  if Offs<=rve.GetOffsBeforeItem(ItemNo) then begin
    dec(ItemNo);
    if ItemNo<0 then
      exit;
    Offs := rve.GetOffsAfterItem(ItemNo);
    Moved := True;    
  end;
  while ItemNo>=0 do begin
    Result := SelectPrevWordInItem(rve, ItemNo, Offs, Moved);
    if Result then
      exit;
    dec(ItemNo);
    if ItemNo<0 then
      exit;
    Offs := rve.GetOffsAfterItem(ItemNo);
    Moved := True;
  end;
end;

{ Selecting next word in the ItemNo-th item of rve. Offs - caret position.
  Moved - if the caret was already moved }
function SelectNextWordInItem(rve: TCustomRichViewEdit; ItemNo, Offs: Integer;
  Moved: Boolean): Boolean;
var ws: WideString;
    WordStartOffs, Len: Integer;
begin
  Result := False;
  if rve.GetItemStyle(ItemNo)<0 then
    exit;
  ws := rve.GetItemTextW(ItemNo);
  Len := Length(ws);
  // skipping delimiters
  while (Offs<=Len) and rve.RVData.IsDelimiterW(Word(ws[Offs])) do begin
    inc(Offs);
    Moved := True;
  end;
  if Offs=Len+1 then
    exit;
  // if the caret was not moved, skipping the current word and delimiters after it
  if not Moved then begin
    while (Offs<=Len) and not rve.RVData.IsDelimiterW(Word(ws[Offs])) do
      inc(Offs);
    if Offs=Len+1 then
      exit;
    while (Offs<=Len) and rve.RVData.IsDelimiterW(Word(ws[Offs])) do
      inc(Offs);
    if Offs=Len+1 then
      exit;
  end;
  // we are at the beginning of word, finding its end
  WordStartOffs := Offs;
  while (Offs<=Len) and not rve.RVData.IsDelimiterW(Word(ws[Offs])) do
    inc(Offs);
  rve.SetSelectionBounds(ItemNo, WordStartOffs, ItemNo, Offs);
  rve.Invalidate;
  Result := True;
end;

{ Selecting the previous word in rve }
function SelectNextWord(rve: TCustomRichViewEdit): Boolean;
var ItemNo, Offs: Integer;
    Moved: Boolean;
begin
  Result := False;
  ItemNo := rve.CurItemNo;
  Offs   := rve.OffsetInCurItem;
  Moved := False;  
  if Offs>=rve.GetOffsAfterItem(ItemNo) then begin
    inc(ItemNo);
    if ItemNo>=rve.ItemCount then
      exit;
    Offs := rve.GetOffsBeforeItem(ItemNo);
    Moved := True;
  end;
  while ItemNo<rve.ItemCount do begin
    Result := SelectNextWordInItem(rve, ItemNo, Offs, Moved);
    if Result then
      exit;
    inc(ItemNo);
    if ItemNo=rve.ItemCount then
      exit;
    Offs := rve.GetOffsBeforeItem(ItemNo);
    Moved := True;
  end;
end;

procedure TForm1.Button1Click(Sender: TObject);
begin
  if not RichViewEdit1.SelectionExists then begin
    RichViewEdit1.SelectCurrentWord;
    Label1.Caption := RichViewEdit1.GetSelTextW;
    if RichViewEdit1.SelectionExists then
      exit;
  end;
  if not SelectPrevWord(RichViewEdit1) then
    Beep;
  Label1.Caption := RichViewEdit1.GetSelTextW;
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
  if not RichViewEdit1.SelectionExists then begin
    RichViewEdit1.SelectCurrentWord;
    Label1.Caption := RichViewEdit1.GetSelTextW;
    if RichViewEdit1.SelectionExists then
      exit;
  end;
  if not SelectNextWord(RichViewEdit1) then
    Beep;
  Label1.Caption := RichViewEdit1.GetSelTextW;   
end;

procedure TForm1.CheckBox1Click(Sender: TObject);
begin
  if FocusedEditor<>nil then
    TRichViewEdit(FocusedEditor).OnCaretMove(FocusedEditor);
  RichViewEdit1.Invalidate;
  RichViewEdit2.Invalidate;  
end;

end.
